#!/usr/bin/env python3



import platform
import sys

import click

from .util.consoleeffects import Colors

try:
    import lxml.etree as ET
except ImportError:
    if platform.system() == 'Windows':
        print('awslogin will not run on your machine yet.  Please follow the instructions at https://github.com/byu-oit/awslogin/releases/tag/lxml to get it running.')
        sys.exit(1)
    else:
        raise


from .util.data_cache import get_status, load_cached_adfs_auth
from .login import cached_login, non_cached_login

__VERSION__ = '0.12.0'

# Enable VT Mode on windows terminal code from:
# https://bugs.python.org/issue29059
# This works not sure if it the best way or not
if platform.system().lower() == 'windows':
    from ctypes import windll, c_int, byref
    stdout_handle = windll.kernel32.GetStdHandle(c_int(-11))
    mode = c_int(0)
    windll.kernel32.GetConsoleMode(c_int(stdout_handle), byref(mode))
    mode = c_int(mode.value | 4)
    windll.kernel32.SetConsoleMode(c_int(stdout_handle), mode)


@click.command()
@click.version_option(version=__VERSION__)
@click.option('-a', '--account', help='Account to login with')
@click.option('-r', '--role', help='Role to use after login')
@click.option('-p', '--profile', default='default', help='Profile to use store credentials. Defaults to default')
@click.option('-s', '--status', is_flag=True, default=False, help='Display current logged in status. Use profile all to see all statuses')
def cli(account, role, profile, status):
    _ensure_min_python_version()

    # Display status and exit if the user specified the "-s" flag
    if status:
        get_status(profile)
        return

    # Use cached SAML assertion if already logged in, else login to ADFS
    adfs_auth_result = load_cached_adfs_auth()
    if adfs_auth_result:
        cached_login(account, role, profile, adfs_auth_result)
        pass
    else:
        non_cached_login(account, role, profile)


def _ensure_min_python_version():
    if not sys.version.startswith('3.6'):
        sys.stderr.write("{}byu_awslogin requires python 3.6{}\n".format(Colors.red, Colors.white))
        sys.exit(-1)


if __name__ == '__main__':
    cli()




import getpass

from .auth.adfs_auth import authenticate
from .auth.assume_role import ask_which_role_to_assume
from .auth.assume_role import assume_role
from .auth.saml import get_account_names, get_saml_assertion
from .util.consoleeffects import Colors
from .util.data_cache import load_last_netid, write_to_config_file, write_to_cred_file, cache_adfs_auth


def cached_login(account, role, profile, adfs_auth_result):
    saml_assertion = get_saml_assertion(adfs_auth_result)

    if not saml_assertion: # Not logged in anymore, so need to re-login
        non_cached_login(account, role, profile)
    else: # Still logged in and got a SAML assertion properly
        _perform_login(adfs_auth_result, saml_assertion, account, role, profile, None)


def non_cached_login(account, role, profile):
    # Get the federated credentials from the user
    net_id, username = _get_user_ids(profile)
    password = getpass.getpass()
    print('')

    ####
    # Authenticate against ADFS with DUO MFA
    ####
    adfs_auth_result = authenticate(username, password)

    # Overwrite and delete the credential variables, just for safety
    username = '##############################################'
    password = '##############################################'
    del username
    del password

    saml_assertion = get_saml_assertion(adfs_auth_result)
    _perform_login(adfs_auth_result, saml_assertion, account, role, profile, net_id)


def _perform_login(adfs_auth_result, saml_assertion, account, role, profile, net_id):
    account_roles_to_assume = _prompt_for_roles_to_assume(saml_assertion, account, role)

    # Assume all requested roles and set in the environment
    for account_role_to_assume in account_roles_to_assume:
        aws_session_token = assume_role(account_role_to_assume, saml_assertion.assertion)

        # If assuming roles across all accounts, then use the account name as the profile name
        if account == 'all':
            profile = account_role_to_assume.account_name

        write_to_cred_file(profile, aws_session_token)
        write_to_config_file(profile, net_id, 'us-west-2', account_role_to_assume.role_name,
                             account_role_to_assume.account_name)
        _print_status_message(account_role_to_assume)

    # By caching the ADFS auth information we can re-assume roles without reauthenticating until the ADFS session expires
    cache_adfs_auth(adfs_auth_result)


def _prompt_for_roles_to_assume(saml_assertion, account, role):
    account_names = get_account_names(saml_assertion)
    account_roles_to_assume = ask_which_role_to_assume(account_names, saml_assertion.get_principal_roles(), account,
                                                       role)
    return account_roles_to_assume


def _print_status_message(assumed_role):
    if assumed_role.role_name == "AccountAdministrator":
        print("Now logged into {}{}{}@{}{}{}".format(Colors.red, assumed_role.role_name, Colors.white,
                                                     Colors.yellow, assumed_role.account_name, Colors.normal))
    else:
        print("Now logged into {}{}{}@{}{}{}".format(Colors.cyan, assumed_role.role_name, Colors.white,
                                                     Colors.yellow, assumed_role.account_name, Colors.normal))


def _get_user_ids(profile):
    # Ask for NetID, or use cached if user doesn't specify another
    cached_netid = load_last_netid(profile)
    if cached_netid:
        net_id_prompt = 'BYU Net ID [{}{}{}]: '.format(Colors.blue,cached_netid,Colors.normal)
    else:
        net_id_prompt = 'BYU Net ID: '
    net_id = input(net_id_prompt) or cached_netid

    # Append the ADFS-required "@byu.local" to the Net ID
    if "@byu.local" in net_id:
        print('{}@byu.local{} is not required'.format(Colors.lblue,Colors.normal))
        username = net_id
    else:
        username = '{}@byu.local'.format(net_id)

    return net_id, username


import base64
from bs4 import BeautifulSoup
import lxml.etree as ET
import requests

default_session_duration = 3600


class SAMLAssertion:
    def __init__(self, assertion):
        self.assertion = assertion

    def get_decoded_assertion(self):
        return ET.fromstring(base64.b64decode(self.assertion))

    def get_principal_roles(self):
        # Find all roles offered by the assertion
        raw_roles = self.get_decoded_assertion().findall(
            './/{*}Attribute[@Name="https://aws.amazon.com/SAML/Attributes/Role"]/{*}AttributeValue'
        )
        aws_roles = [element.text.split(',') for element in raw_roles]

        # Note the format of the attribute value is provider_arn, role_arn
        principal_roles = [role for role in aws_roles if ':saml-provider/' in role[0]]
        return principal_roles

    def get_session_duration(self):
        aws_session_duration = default_session_duration
        # Retrieve session duration
        for element in self.get_decoded_assertion().findall(
                './/{*}Attribute[@Name="https://aws.amazon.com/SAML/Attributes/SessionDuration"]/{*}AttributeValue'):
            aws_session_duration = int(element.text)
        return aws_session_duration


def get_saml_assertion(adfs_auth_result):
    headers = {
        'Accept-Language': 'en',
        'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko',
        'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
        'Accept': 'text/plain, */*; q=0.01',
    }

    roles_page_response = adfs_auth_result.session.post(
        adfs_auth_result.action_url,
        verify=True,
        headers=headers,
        allow_redirects=True,
        data={
          'AuthMethod': 'DuoAdfsAdapter',
          'Context': adfs_auth_result.context,
          'sig_response': adfs_auth_result.signed_response
        }
    )

    if roles_page_response.status_code != 200:
        raise RuntimeError(
            u'Issues during redirection to aws roles page. The error response {}'.format(
                roles_page_response
            )
        )

    roles_page_html_soup = BeautifulSoup(roles_page_response.text, 'lxml')
    return _extract_saml_assertion(roles_page_html_soup)


def get_account_names(saml_assertion):
    saml_url = "https://signin.aws.amazon.com:443/saml"
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'
    }
    response = requests.post(saml_url, headers=headers, data={
        'SAMLResponse': saml_assertion.assertion
    })
    response.raise_for_status()
    html_response = ET.fromstring(response.text, ET.HTMLParser())
    account_names = {}
    for element in html_response.findall('.//div[@class="saml-account-name"]'):
        account_id = element.text.split(' ')[2].replace('(', '').replace(')', '')
        account_name = element.text.split(' ')[1]
        account_names[account_id] = account_name

    return account_names


def _extract_saml_assertion(saml_page_html_soup):
    hidden_form = saml_page_html_soup.find('form', {"name": "hiddenform"})
    if not hidden_form:
        return None

    # Retrieve Base64-encoded SAML assertion from form SAMLResponse input field
    assertion = hidden_form.find('input', {"name": "SAMLResponse"})['value']

    # If we did not get an error, but also do not have an assertion,
    # then the user needs to authenticate
    if not assertion:
        print("Unknown error: No assertion returned")
        exit(1)

    return SAMLAssertion(assertion)
